% parametric squared euclidean distance
% with squared diagonal metric 
classdef ParametricMatrixDistance < handle
    properties (GetAccess='public',SetAccess = 'private')
        bInitialized = false;
        Omega       = [];
        Lambda      = [];
        Quadratic   = true;
        D           = 0;
        subdim      = 0;
        InitialLambda = [];
    end
    
    methods
        % constructor
        function obj = ParametricMatrixDistance(NrDimensions,subdim)
            obj        = obj@handle();
            obj.Quadratic = false;
            obj.subdim    = subdim;
            if(~exist('subdim','var') || subdim == NrDimensions)
                obj.Quadratic = true;
                obj.subdim    = NrDimensions;
            end
            if(0)%obj.Quadratic)
                obj.Omega = eye(NrDimensions)/sqrt(NrDimensions);
            else
                obj.Omega = 2*rand(subdim,NrDimensions) - 1;
                obj.Omega = obj.Omega ./ sqrt(sum(diag(obj.Omega'*obj.Omega)));
            end
            obj.D            = NrDimensions;            
            obj.bInitialized = true;
            obj.Lambda = obj.Omega'*obj.Omega;
            obj.InitialLambda = obj.GetLambda();
        end
        
        function Ret = GetOmega(me)
            Ret = me.Omega;
            return
        end
        
        
        function Ret = GetLambda(me)
            Ret = me.Lambda;
            return
        end        
    
        function Ret = GetParameter(me)
            Ret = me.GetOmega(); % this version returns the full Omega matrix as parameters
            Ret = Ret(:);
        end        

        function [Ok] = SetParameter(me,Omega)
            if(isvector(Omega))
                assert(numel(Omega)==[me.subdim* me.D])
                Omega = reshape(Omega,me.subdim, me.D);
            else                
                assert(all(size(Omega)==[me.subdim me.D]))
            end
            if(1) % to check gradient
                me.Omega = Omega ./ sqrt(sum(diag(Omega'*Omega)));                
            else
                me.Omega = Omega;
            end
            me.Lambda = (me.Omega'*me.Omega);
            Ok = 1;
        end
        
        function Value = subsref(me,Indices)
            Value = [];
            NumInd = numel(Indices);
            if(NumInd==1) % direct access 
                Subs       = Indices.subs;
                NumIndSubs = numel(Subs);
                switch Indices.type
                    case '()'     % indexing                        
                        if(NumIndSubs==2)
                            Value = me.CalculateDistance(Subs{1},Subs{2});
                        else
                            ME = MException('Class:ParametricDistance:Error','number of parameters must be two');
                            throw(ME);
                        end
                    case '.'     % class member
                        if(strcmp(Subs,'Lambda'))
                            Value = me.GetLambda();
                        else
                            Value = me.(Subs);
                        end
                end
            else % should be a function call
                if(strncmp(Indices(1).type,'.',1))   
                    Value = builtin('subsref', me, Indices); 
                end
            end
        end       
        
        % input D x N, D x M with D number of dimensions equal number of columns of Omega
        function Dist = CalculateDistance(me,X,Y)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end            
            N             = size(X,1);
            iNrPrototypes = size(Y,1);
            weucHelp = @(Diff)reshape(sum(Diff' .* (me.Lambda * Diff)',2),N,iNrPrototypes); % squared mahalanobis distance (non-squared weights)
            DistHelp = @(X,Y)weucHelp(reshape(permute(permute(repmat(Y, [1,1,N]),[3 2 1]) - repmat(X,[1,1,iNrPrototypes]),[2 1 3]),me.D,N*iNrPrototypes) ); 
            Dist = DistHelp(X,Y);
        end
        
        function Dist = CalculateDistanceDerivativeLambdaMulti(me,X,Y,Degree,varargin)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end
            
            N             = size(X,1);

            switch Degree
                case 1
                    DistHelp2 = @(DiffScale,Diff)reshape(permute(permute(repmat(DiffScale, [1,1,me.D]), [3 2 1]) .* repmat(Diff,[1,1,me.subdim]),[3 1 2]),me.subdim * me.D,N);
                    weucHelp = @(Diff)DistHelp2(me.Omega * Diff,Diff); % squared euclidean distance with squared (!) weights            
                    Dist = 2 .* weucHelp(bsxfun(@minus,X,Y)');
                otherwise
                    throw(MException('Class:ParametricDistance:Error','only first  derivative supported'));                    
            end
        end        
        
        function Dist = CalculateDistanceDerivative2ndArgMultiXY(me,X,Y)       
            if(size(X,1) == me.D && size(X,1) ~= size(X,2)) 
                X = X';
            end
            
            if(size(Y,1) == me.D && size(Y,1) ~= size(Y,2)) 
                Y = Y';
            end        
            N             = size(X,1);
            iNrPrototypes = size(Y,1);
            weucHelp = @(Diff)me.Lambda * Diff; % squared euclidean distance with squared (!) weights            
            DistHelp = @(X,Y)weucHelp(reshape(permute(permute(repmat(Y, [1,1,N]),[3 2 1]) - repmat(X,[1,1,iNrPrototypes]),[2 1 3]),me.D,N*iNrPrototypes) ); 
            Dist = 2 .* permute(reshape(DistHelp(X,Y),[me.D,N,iNrPrototypes]),[2 1 3]);
        end                
        
    end
end