% mD - normalized data
% L  - label vector
% v  - number of folds
function [vModelResults,vClassifyResults] = crossvalidation_glvq(mD, L, v,varargin)
%CROSSVAL K-fold cross validation.

error(nargchk(3, 6, nargin))
[N] = size(mD,1);
if nargin < 2
  v = [];
elseif nargin <4
    varargin = [];
end

if isempty(v)
  V = 1:N;
  v = N;
elseif length(v) == 1
  if ~isa(v, 'double') | ~isreal(v) | v <= 0 | round(v) ~= v
    error('V must be a positive, non-zero integer.')
  elseif v > N
    error('V must be less than the number of observations.')
  elseif v == N
    V = 1:N;
  else
    w = v*fix(N/v);
    r = randperm(v);
    V(randperm(N)) = [reshape(repmat(1:v, w/v, 1), w, 1); ...
		      r(1:N-w)'];
  end
else
  if ~isa(v, 'double') | ~isreal(v) | prod(size(v)) ~= length(v) | ...
	any(round(v) ~= v | v <= 0 | isinf(v))
    error(['Cross-validation set indeces V must be a vector of' ...
	   ' positive, finite, non-zero integers.'])
  elseif length(v) ~= N
    error(['Length of set indeces V must be same as number of' ...
	   ' observations in X and K.'])
  elseif ~all(sum(sparse(1:N, v, 1)))
    error('Set indeces V may not have empty groups.')
  end
  
  V = v;
  v = max(V);
end
  
c = zeros(N, 1);
vModelResults    = [];
vClassifyResults = [];
  
Label                  = L(:);
vLabels                = unique(L);
NrLabels               = numel(vLabels);

  
for i = 1:v
  fprintf('*')
  s = warning;
  warning off
  
  % determine training model
  CurrentSet      = find(V ~= i);
  RSamples        = randsample(numel(CurrentSet),min(numel(CurrentSet),max(100,0.01*numel(CurrentSet))));

  iCycles       = 100;
  [NTemp,C]     = size(L);
  if(C==1)
    iProtos     = numel(unique(L));
  else
    iProtos     = C;
  end

  tStart        = cputime;
  
  if(isempty(varargin))      
      [oModel]      = glvq_train(mD(CurrentSet,:),Label(CurrentSet),iProtos);
  else
      iProtos       = varargin{1}{3};      
      [oModel]      = glvq_train(mD(CurrentSet,:),Label(CurrentSet),iProtos, varargin{1}{:});
  end
  estimatedLabels   = glvq_classify(mD(CurrentSet,:), oModel);
  dAccTrain         = mean(Label(CurrentSet) == estimatedLabels(:));

  dRunTime      = cputime-tStart; 

   % classify test data
  CurrentTestSet = find(V == i); 
  
  estimatedLabels = glvq_classify(mD(CurrentTestSet,:), oModel);
  dAccTest        = mean(Label(CurrentTestSet) == estimatedLabels(:));
  
  sModel = struct('Models',oModel,'Labels',{vLabels},'V',{V},'dRunTime',dRunTime,'AccTrain',dAccTrain,'dAccTest',dAccTest);  
  warning(s);
  
  vClassifyResults = [vClassifyResults dAccTest];
  vModelResults    = [vModelResults sModel];  
end
