function [f G]  = glvq_optfun(variables,Data, fDist, LabelEqualsPrototype,varargin)
% [f G] = glvq_optfun(variables) 
% function to be optimized with generalized learning vector quantization 
%
% Input:
%       variables            = see glvq
%       fDist                = function pointer to distance calculation class which also provides gradients
%       LabelEqualsPrototype = N x M bool matrix - where prototype labels match datapoint labels
%
% Output:
%       f                    = cost function error current model
%       G                    = gradient
%
% Conditions of GNU General Public License, version 2 apply.
% See file 'license-gpl2.txt' enclosed in this package.
% Programs are not for use in critical applications!
% 
% (C) Frank-Michael Schleif
% University of Bielefeld
% fschleif@techfak.uni-bielefeld.de
%
bGradientCheckOff = false;
[N,nb_prototypes] = size(LabelEqualsPrototype);
D                 = fDist.D;
Neurons           = reshape(variables(1:nb_prototypes*D), nb_prototypes,D);
bMetricUpdate     = false;
bMatrixUpdate     = false;
if(numel(variables)>nb_prototypes*D)
    bMetricUpdate     = true;
    MetricParameters  = variables(nb_prototypes*D+1:end);
    m = 1;
    if(isprop(fDist,'Lambda'))
        bMatrixUpdate = true;
        if(isprop(fDist,'m'))
            m = fDist.m;
        elseif(isprop(fDist,'subdim'))
            m = fDist.subdim;
        else
            bMatrixUpdate = false;
        end
    end
    fDist.SetParameter(MetricParameters);
end
mDist             = fDist.CalculateDistance(Data,Neurons); 

distwrong   = zeros(1,N);
pidxwrong   = zeros(1,N);
distcorrect = zeros(1,N);
pidxcorrect = zeros(1,N);

for(k=1:N)
    vValidProtos = find(LabelEqualsPrototype(k,:));
    if(~isempty(vValidProtos))
        [distcorrect(k) idx] = min(mDist(k,vValidProtos));
        pidxcorrect(k) = vValidProtos(idx);
    end
    vInValidProtos = find(~LabelEqualsPrototype(k,:));
    if(~isempty(vInValidProtos))
        [distwrong(k) idx] = min(mDist(k,vInValidProtos));
        pidxwrong(k)= vInValidProtos(idx);
    end
end

distcorrectpluswrong = distcorrect + distwrong;
distcorrectminuswrong = distcorrect - distwrong;
mu = distcorrectminuswrong ./ distcorrectpluswrong;
f  = sum(mu);
if(0)
    hold on
    PlotData(Data(:,[3,4]),varargin{1}{1});
    PlotData2(Neurons(:,[3,4]),varargin{1}{2});
    pause(1)
    hold off;
    close all;
end
if nargout > 1  % gradient needed not just function eval
    G = zeros(size(variables)); % initially no gradient
        
        % weight vector update
        TauP    = bsxfun(@rdivide,2*distwrong,max(realmin,(distcorrectpluswrong.^2)));
        TauM    = bsxfun(@rdivide,2*distcorrect,max(realmin,(distcorrectpluswrong.^2)));
        GradN2D = fDist.CalculateDistanceDerivative2ndArgMultiXY(Neurons,Data);
        NeuronsUpdate = zeros(nb_prototypes,D);
        
        if(bGradientCheckOff) % to check gradient this needs to be off
            iStepSize = ceil(N/50);
        else
            iStepSize = 1;
        end
        for(i=1:iStepSize:N)
            if(pidxcorrect(i)>0)
                NeuronsUpdate(pidxcorrect(i),:) = NeuronsUpdate(pidxcorrect(i),:) + TauP(i) * GradN2D(pidxcorrect(i),:,i);
            end
            if(pidxwrong(i)>0)
                NeuronsUpdate(pidxwrong(i),:)   = NeuronsUpdate(pidxwrong(i),:)   - TauM(i) * GradN2D(pidxwrong(i),:,i);
            end
        end
        G = -NeuronsUpdate(:);
                
        vUsedTrainingSamplesIndex = 1:iStepSize:N;        
        iUsedTrainingSamples      = numel(vUsedTrainingSamplesIndex);
        % metric update
        if(bMetricUpdate)
            LambdaUpdate    = zeros(numel(MetricParameters),1);
            TempDerivLambda = zeros(nb_prototypes,iUsedTrainingSamples,numel(MetricParameters));
            for(p=1:nb_prototypes)
                TempDerivLambda(p,:,:) = fDist.CalculateDistanceDerivativeLambdaMulti(Data(vUsedTrainingSamplesIndex,:),Neurons(p,:),1)';
            end
            for(i=1:iUsedTrainingSamples)
                iCurrentSample = vUsedTrainingSamplesIndex(i);
                if(pidxcorrect(iCurrentSample)>0)
                    LambdaUpdate = LambdaUpdate + (TauP(iCurrentSample) * reshape(TempDerivLambda(pidxcorrect(iCurrentSample),i,:),D*m,1)')';                                
                end
                if(pidxwrong(iCurrentSample)>0)
                    LambdaUpdate = LambdaUpdate - (TauM(iCurrentSample) * reshape(TempDerivLambda(pidxwrong(iCurrentSample),i,:),D*m,1)')';                                
                end

            end
            G = [G; LambdaUpdate(:)];
            if(~isLegal(G(:)))
                pause;
            end
        end        
end
